# This folder cannot be nested until the import needs of the base packages and
# the cffi extra are more differentiated. Right now, the cffi extra needs only
# the setuptools package and only in Python 3.12.
